import json
import os
from pathlib import Path
from typing import Dict, List

from tftokenizers.file import save_json
from tftokenizers.types import TokenizerAttributes

DATABASE_PATH = Path("database.json")
TOKENIZER_ENTRY = "tokenizers"
TOKENIZER_NAME_ATTR = "name_or_path"
EXECUTE_PERMISSION = 0o0755


def create_database(database_path=DATABASE_PATH) -> None:
    """Save the files of a Huggingface tokenizer locally."""
    database: Dict[str, List[str]] = {TOKENIZER_ENTRY: []}
    with open(database_path, "w+", encoding="utf-8") as f:
        save_json(database, f)


def get_database():
    with open(DATABASE_PATH, "r+", encoding="utf-8") as f:
        data = json.load(f)
    return data


def insert_database_entry(attributes: TokenizerAttributes):
    database: dict = get_database()
    d = database[TOKENIZER_ENTRY]
    new_entry = attributes.dict()
    is_found = False

    for idx, entry in enumerate(d):
        if entry[TOKENIZER_NAME_ATTR] == new_entry[TOKENIZER_NAME_ATTR]:
            is_found = True
            database[TOKENIZER_ENTRY][idx] = new_entry
            break

    if not is_found:
        database[TOKENIZER_ENTRY].append(new_entry)

    return database


def save_tokenizer_attributes(attributes: TokenizerAttributes):
    if not os.path.isfile(DATABASE_PATH):
        create_database()

    with open(DATABASE_PATH, "r+", encoding="utf-8") as f:
        # data = json.load(f)
        data = insert_database_entry(attributes)
        f.seek(0)
        save_json(data, f)


def create_download_script():
    """Create bash script to download all tokenizers in database."""
    database = get_database()

    SCRIPT_NAME = "scripts/download_tokenizers.sh"
    MODULE = "tftokenizers"
    FILE = "download.py"

    with open(SCRIPT_NAME, "w") as f:
        f.writelines("#!/usr/bin/env bash\n\n")

        for entry in database[TOKENIZER_ENTRY]:
            tokenizer_name = entry[TOKENIZER_NAME_ATTR]
            f.writelines(f"python {MODULE}/{FILE} -n {tokenizer_name}\n")

    os.chmod(SCRIPT_NAME, EXECUTE_PERMISSION)



from typing import Any, Dict, List, Optional, Tuple

import tensorflow as tf

from tftokenizers.types import PaddingStrategies, TemplateType


def map_special_tokens_to_ids(
    vocab: List[str], special_tokens_or_config: Dict[str, str]
) -> Dict[str, tf.Tensor]:
    """Map special tokens ([SEP]) to their token id as a tensor."""
    special_tokens_map: Dict[str, tf.Tensor] = {}

    for name, symbol in special_tokens_or_config.items():
        if "_token" not in name or "_tokens" in name:
            continue
        if isinstance(symbol, str):
            _token_id = vocab.index(symbol)
            token_id: tf.Tensor = to_tensor(_token_id)
            special_tokens_map[symbol] = token_id

    return special_tokens_map


def num_special_tokens(proceesing_type: TemplateType, config: Dict[str, Any]):
    r"""Sets the number of special tokens when processing a sentence.

    Configuration is based on the `post_processor` template from \
        the configurations in Huggingface Tokenizers, \
        The template specifies which tokens and in what order \
        special tokens should be added to a single or pair os sentences.
    """

    assert proceesing_type in ["single", "pair"]

    template: List[Dict[str, Any]] = config["post_processor"][proceesing_type]
    return len([s for s in template if s.get("SpecialToken") is not None])


def parse_args_to_keywords(args) -> Tuple[str, str]:
    tokenizer_name: str = args.name_or_path
    save_name: str = tokenizer_name.replace("/", "_")
    path = f"{args.output_dir}/{save_name}"
    return tokenizer_name, path


def to_tensor(x: int) -> tf.Tensor:
    return tf.constant(x, dtype=tf.int64)


def list_to_tensor(lst: List[int]):
    return [to_tensor(num) for num in lst]


def set_valid_max_seq_length(
    desired=None, model_max_length=512, num_max_tokens_in_seq=510
) -> int:
    """Calculate and set the maximum allowed sequence length."""
    min_length = model_max_length - num_max_tokens_in_seq
    max_length = num_max_tokens_in_seq

    if desired is None or desired > max_length or desired <= min_length:
        desired = max_length
    return desired


def set_valid_padding(padding: Optional[PaddingStrategies]) -> PaddingStrategies:
    padding = padding if padding is not None else PaddingStrategies.LONGEST
    return padding



"""Help identify models, class type between a Huggingface tokenizer and Tensorflow implementation."""

import tensorflow as tf
import tensorflow_text as text
from transformers import PreTrainedTokenizer


def get_tokenizer_type(model_name_or_path, path, **kwargs):
    pass


# TODO: could be enum type
def find_tf_base_tokenizer(tokenizer_type, **kwargs):
    pass


def load_tokenizer(tokenizer_tf_base, tokenizer_config_params, **kwargs):
    pass


def detect_and_load_tokenizer(tokenizer: PreTrainedTokenizer, path: str, **kwargs):
    # TODO: Should load tokenizer based on `tokenizer.backend_tokenizer.model`
    """Load the corresponging base tokenizer from Huggingaface and map it to a Tensorflow model.

    We set lower_case=False, since we want the tokenizer vocabulary to be similarly cased to Huggingface
    """
    vocab_file = tokenizer.vocab_files_names["vocab_file"]
    vocab_path = f"{path}/{vocab_file}"

    """
    (vocab_lookup_table: Unknown,
    suffix_indicator: Unknown = "##",
    max_bytes_per_word: Unknown = 100,
    max_chars_per_token: Unknown = None,
    token_out_type: Unknown = dtypes.int64,
    unknown_token: Unknown = "[UNK]",
    split_unknown_characters: Unknown = False,
    lower_case: Unknown = False,
    keep_whitespace: Unknown = False,
    normalization_form: Unknown = None,
    preserve_unused_token: Unknown = False,
    basic_tokenizer_class: Unknown = BasicTokenizer) -> None
    """
    return text.BertTokenizer(
        vocab_path,
        token_out_type=tf.int64,
        lower_case=tokenizer.do_lower_case,
        **kwargs,
    )



from typing import Dict, List, Union

import tensorflow as tf
import tensorflow.keras as keras
from transformers import AutoTokenizer, TFAutoModel
from transformers.utils.logging import set_verbosity_error

from tftokenizers.tokenizer import TFAutoTokenizer

set_verbosity_error()
tf.get_logger().setLevel("ERROR")


def compare_models(pretrained_model_name: str, inputs: Union[str, List[str]]):
    """Helper function for comparing TF and HF tokenizers."""

    model = TFAutoModel.from_pretrained(pretrained_model_name)

    tf_tokenizer = TFAutoTokenizer.from_pretrained(pretrained_model_name)
    tf_hub_model = TFModel(model=model, tokenizer=tf_tokenizer)
    tf_output = tf_hub_model(inputs, training=False)

    hf_tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name)
    hf_tokens: Dict[str, tf.Tensor] = hf_tokenizer.batch_encode_plus(
        inputs,
        padding=True,
        truncation=True,
        return_token_type_ids=False,
        return_tensors="tf",
    )
    hf_output = model(hf_tokens)[0]  # "last_hidden_state"

    return {"tf_output": tf_output, "hf_output": hf_output}


class TFModel(keras.layers.Layer):
    def __init__(self, model, tokenizer, **kwargs):
        super(TFModel, self).__init__(**kwargs)
        self.tokenizer = tokenizer
        self.model = model

    def call(
        self,
        inputs,
        verbose=False,
        training=True,
    ):
        tokens = self.tokenizer.tokenize(inputs=inputs)
        if verbose:
            print(tokens)

        # we only keep tensor 0 (last_hidden_state) for now
        embeddings = self.model(tokens, training=training)
        last_hidden_out = embeddings[0]
        return last_hidden_out



import argparse
import logging

from rich.logging import RichHandler
from transformers import AutoTokenizer

from tftokenizers.database import save_tokenizer_attributes
from tftokenizers.types import TokenizerAttributes
from tftokenizers.utils import parse_args_to_keywords


def parse_args():
    parser = argparse.ArgumentParser(
        description="Save a huggingface tokenizer for conversion to a Tensorflow Text tokenizer."
    )
    parser.add_argument(
        "--name_or_path",
        "-n",
        type=str,
        help="Name to pretrained model or tokenizer from huggingface.co/models.",
        default="bert-base-uncased",
        # required=True
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="saved_tokenizers",
        help="Where to store the Huffingface tokenizer configs.",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Display debug info if passed in as argument",
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    logging.basicConfig(
        format="%(message)s",
        datefmt="[%X]",
        level="NOTSET" if args.verbose else logging.INFO,
        handlers=[RichHandler(rich_tracebacks=True)],
    )

    log = logging.getLogger("rich")
    log.info("Converting tokenizer...")

    name, path = parse_args_to_keywords(args)
    tokenizer = AutoTokenizer.from_pretrained(name)
    tokenizer.save_pretrained(path)
    log.info(f"Saved tokenizer to: {path}")

    attributes = TokenizerAttributes(
        name_or_path=name,
        export_path=path,
    )
    save_tokenizer_attributes(attributes)
    log.info("Saving a download script")


if __name__ == "__main__":
    main()



import glob
import json
import pathlib
from enum import Enum
from io import TextIOWrapper
from typing import List, Tuple


class FileTypes(str, Enum):
    TXT = "txt"
    JSON = "json"
    NONE = "None"


def get_filename_from_path(path: str, filename: str) -> str:
    """Find a file name prefix in a folder."""
    if filename in path:
        path_without_filename = "/".join(path.split("/")[:-1])
        path = path_without_filename

    for file_path in glob.glob(f"{path}/*"):
        file = file_path.split("/")[-1]

        if filename in file and filename not in path:
            filename = file
            break

    file_path = f"{path}/{filename}"
    return file_path


def get_filename_and_type_from_path(path: str, filename: str) -> Tuple[str, str]:
    """Find a file name prefix in a folder."""
    file_type = "None"

    if filename in path:
        path_without_filename = "/".join(path.split("/")[:-1])
        path = path_without_filename

    for file_path in glob.glob(f"{path}/*"):
        file = file_path.split("/")[-1]
        file_type = file.split(".")[-1]

        if filename in file and filename not in path:
            filename = file
            break

    file_path = f"{path}/{filename}"
    return file_path, file_type


def get_vocab_from_path(path: str) -> List[str]:

    vocab: List[str]
    filename = "vocab"
    file_path, filetype = get_filename_and_type_from_path(path, filename)

    if filetype == FileTypes.TXT:
        vocab = parse_and_load_txt_vocab(file_path)
    elif filetype == FileTypes.JSON:
        vocab = parse_and_load_json_vocab(file_path)
    else:
        raise ValueError(f"Could not parse the given file/path: {path}")
    return vocab


def load_txt(file_path: str):
    """Loads txt file from path."""
    data = pathlib.Path(file_path).read_text().splitlines()
    return data


def load_json(file_path: str):
    """Loads json file from path."""
    with open(file_path) as f:
        data = json.load(f)
    return data


def save_json(data: dict, f: TextIOWrapper):
    return json.dump(data, f, ensure_ascii=False, indent=4)


def parse_and_load_txt_vocab(file_path: str) -> List[str]:
    """Parses tokenizer vocab from a txt file."""
    return load_txt(file_path)


def parse_and_load_json_vocab(file_path: str) -> List[str]:
    """Parses tokenizer vocab from a json file."""
    data = load_json(file_path)
    return [key for key in data]



from enum import Enum
from typing import Any, Dict, Tuple, Union

import tensorflow as tf
from pydantic import BaseModel

SPECIAL_TOKENS_DICT = Dict[str, Dict[str, Union[str, int, Any]]]
PROCESSING_STEP = Dict[str, Dict[str, str]]
TF_TENSORS = Union[tf.RaggedTensor, tf.SparseTensor, tf.IndexedSlices, tf.Tensor, Any]
TF_TOKENIZER_INPUT = Tuple[TF_TENSORS, TF_TENSORS]


class PaddingStrategies(str, Enum):
    """Define how input in the tokenizer should be padded."""

    MAX_LENGTH = "max_length"
    LONGEST = "longest"
    DO_NOT_PAD = "do_not_pad"
    NONE = "None"


class SpecialTokens(str, Enum):
    """Name of special tokens a particular tokenizer may have."""

    UNK_TOKEN = "unk_token"
    END_TOKEN = "eos_token"
    PAD_TOKEN = "pad_token"
    START_TOKEN = "bos_token"
    SEP_TOKEN = "sep_token"
    CLS_TOKEN = "cls_token"
    MASK_TOKEN = "mask_token"


class TemplateType(str, Enum):
    """Options for how to process a `post_processor` template."""

    SINGLE = "single"
    PAIR = "pair"


class TokenizerAttributes(BaseModel):
    name_or_path: str
    export_path: str


class NotValidPaddingStrategy(Exception):
    """Exception raised for not supported padding strategy.

    padding strategy for a tokenizer, currently supports strategies from `PaddingStrategies`:class
    """

    def __init__(
        self,
        padding,
        message="The padding strategy chosen is not supported. \
            Use one listed in the class `PaddingStrategies`.",
    ):
        self.strategy = padding
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"Chosen padding strategy: {self.strategy}\n{self.message}"



import tensorflow as tf
import tensorflow_text as text
from transformers import AutoTokenizer, TFAutoModel
from transformers.utils.logging import set_verbosity_error

from tftokenizers.file import get_filename_from_path, get_vocab_from_path, load_json
from tftokenizers.model import TFModel
from tftokenizers.tokenizer import TFAutoTokenizer, TFTokenizerBase

set_verbosity_error()
tf.get_logger().setLevel("ERROR")

if __name__ == "__main__":
    # Load base models from Huggingface
    pretrained_model_name = "bert-base-cased"
    pretrained_model_name = "distilbert-base-cased"
    pretrained_model_name = "distilbert-base-multilingual-cased"
    pretrained_model_name = "bert-base-multilingual-cased"
    pretrained_model_name = "KB/bert-base-swedish-cased"

    tf_tokenizer = TFAutoTokenizer.from_pretrained(pretrained_model_name)
    hf_tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name)
    model = TFAutoModel.from_pretrained(pretrained_model_name)

    # Three ways to load the TF tokenizer
    # a) by model_name
    tf_tokenizer = TFAutoTokenizer.from_pretrained(pretrained_model_name)

    # b) bundled with the model, similar to TFHub
    model = TFAutoModel.from_pretrained(pretrained_model_name)
    custom_model = TFModel(model=model, tokenizer=tf_tokenizer)

    # c) from vocab file
    PATH = "saved_tokenizers/bert-base-uncased"
    vocab = get_vocab_from_path(PATH)
    vocab_path = get_filename_from_path(PATH, "vocab")

    config = load_json(f"{PATH}/tokenizer_config.json")
    tokenizer_spec = load_json(f"{PATH}/tokenizer.json")
    special_tokens_map = load_json(f"{PATH}/special_tokens_map.json")

    tokenizer_base_params = dict(lower_case=True, token_out_type=tf.int64)
    tokenizer_base = text.BertTokenizer(vocab_path, **tokenizer_base_params)
    custom_tokenizer = TFTokenizerBase(
        vocab_path=vocab_path,
        tokenizer_base=tokenizer_base,
        hf_spec=tokenizer_spec,
        config=config,
    )

    # Usage
    # TF tokenizer have the same interface as HF transformers
    sentence = "Huggingface to Tensorflow tokenizers"
    print(tf_tokenizer.batch_encode_plus([sentence])["input_ids"])
    print(
        hf_tokenizer.batch_encode_plus(
            [sentence],
            padding=True,
            truncation=True,
            return_token_type_ids=False,
            return_tensors="tf",
        )["input_ids"]
    )

    # Create a Reusable SavedModel
    custom_model = TFModel(model=model, tokenizer=tf_tokenizer)

    # With the tokenizer in Tensorflow, it can handle Tensors and python string
    string_tf = tf.constant(["Hello from Tensorflow"])
    output = custom_model(string_tf)

    # And pass input for either inference (Default) or training
    output = custom_model(string_tf, training=False)  # inference
    output = custom_model(string_tf, training=True)  # training

    # Or pass a batch of sentences, similar to Huggingface
    s1 = "sponge bob squarepants is an avenger"
    s2 = "Huggingface to Tensorflow tokenizers"
    s3 = "HelLo!"
    output = custom_model(
        [s1, s2, s3],
    )
    print(output)



"""Use Huggingface Transformer and Tokenizers as Tensorflow Resuable SavedModels."""

__version__ = "0.1.6"

from .detect import detect_and_load_tokenizer as detect_and_load_tokenizer
from .detect import find_tf_base_tokenizer as find_tf_base_tokenizer
from .detect import get_tokenizer_type as get_tokenizer_type
from .detect import load_tokenizer as load_tokenizer
from .file import get_vocab_from_path as get_vocab_from_path
from .model import TFModel as TFModel
from .tokenizer import TFAutoTokenizer as TFAutoTokenizer
from .tokenizer import TFTokenizerBase as TFTokenizerBase



import re
from typing import Any, Dict, List, Optional, Union

import tensorflow as tf
import tensorflow_text as text
from transformers import AutoTokenizer, PreTrainedTokenizer

from tftokenizers.detect import detect_and_load_tokenizer
from tftokenizers.file import get_vocab_from_path, load_json
from tftokenizers.types import (
    PROCESSING_STEP,
    TF_TENSORS,
    TF_TOKENIZER_INPUT,
    NotValidPaddingStrategy,
    PaddingStrategies,
    SpecialTokens,
    TemplateType,
)
from tftokenizers.utils import (
    list_to_tensor,
    map_special_tokens_to_ids,
    num_special_tokens,
    set_valid_max_seq_length,
    set_valid_padding,
)


def compare_tokenizers(pretrained_model_name: str, inputs: Union[str, List[str]]):
    """Helper function for comparing TF and HF tokenizers."""

    tf_tokenizer = TFAutoTokenizer.from_pretrained(pretrained_model_name)
    hf_tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name)

    tf_tokens = tf_tokenizer.batch_encode_plus(inputs)
    hf_tokens = hf_tokenizer.batch_encode_plus(
        inputs,
        padding=True,
        truncation=True,
        return_token_type_ids=False,
        return_tensors="tf",
    )

    return {"tf_tokens": tf_tokens, "hf_tokens": hf_tokens}


def save_and_load_tokenizer_to_path(
    tokenizer: PreTrainedTokenizer,
    model_name_or_path: str,
    output_dir="saved_tokenizers",
):
    path = model_name_or_path
    if output_dir not in model_name_or_path:
        path = f"{output_dir}/{model_name_or_path}"
        tokenizer.save_pretrained(path)
    return path


class TFAutoTokenizer(tf.Module):
    def __init__(
        self,
        model_name_or_path,
        max_length: Optional[int] = None,
        padding: Optional[PaddingStrategies] = None,
        output_dir="saved_tokenizers",
        tokenizer_config_params: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        super(TFAutoTokenizer, self).__init__(**kwargs)
        config: PreTrainedTokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        path = save_and_load_tokenizer_to_path(config, model_name_or_path, output_dir)
        pipeline = load_json(f"{path}/tokenizer.json")

        self.tokenizer_file = config.vocab_files_names["tokenizer_file"]
        self.vocab_file = config.vocab_files_names["vocab_file"]
        # TODO: `vocab_file` Could be vocab or SentencePieice model
        vocab_path = f"{path}/{self.vocab_file}"
        self.vocab_path = vocab_path
        self.vocab = get_vocab_from_path(vocab_path)
        self.vocab_size = config.vocab_size
        self.vocab_dict = config.vocab
        self._vocab = tf.Variable(self.vocab)

        # TODO: Load in multiple steps
        self.tokenizer = detect_and_load_tokenizer(config, path)
        # self.tokenizer_type = get_tokenizer_type(model_name_or_path, path)
        # self.tokenizer_tf_base = find_tf_base_tokenizer(self.tokenizer_type)
        # self.tokenizer = load_tokenizer(
        #     self.tokenizer_tf_base, tokenizer_config_params, **kwargs
        # )

        self.special_tokens = map_special_tokens_to_ids(
            self.vocab, load_json(f"{path}/tokenizer_config.json")
        )
        self.all_special_tokens = config.all_special_tokens
        self.all_special_ids = config.all_special_ids
        self.all_special_ids_tf = list_to_tensor(config.all_special_ids)

        # TODO: Should be determined by the tokenizer config
        # POtential solution could be:
        # $ c = tokenizer.backend_tokenizer.post_processor
        # $ sentences_pair  = model_max_length - c.num_special_tokens_to_add(is_pair=False)
        # $ single_sentence = model_max_length - c.num_special_tokens_to_add(is_pair=True)
        self.max_len_sentences_pair = 509
        self.max_len_single_sentence = 510
        self.model_max_length = max_length if max_length is not None else 510

        self.model_input_names = config.model_input_names
        self.name_or_path = config.name_or_path
        self.saved_at_path = path

        # Tokenizer pipeline
        self.decoder = pipeline["decoder"]
        self.model = pipeline["model"]
        self.normalizer = pipeline["normalizer"]
        self.post_processor = pipeline["post_processor"]
        self.pre_tokenizer = pipeline["pre_tokenizer"]

        self.decoder_type = pipeline["decoder"]["type"]
        self.model_type = pipeline["model"]["type"]
        self.normalizer_type = pipeline["normalizer"]["type"]
        self.post_processor_type = pipeline["post_processor"]["type"]
        self.pre_tokenizer = pipeline["pre_tokenizer"]["type"]

        self.padding = padding if padding is not None else PaddingStrategies.LONGEST
        # self.max_length = (
        #    max_length if max_length is not None else self.model_max_length
        # )
        self.max_length = 512
        self._reserved_tokens = list(self.special_tokens.keys())
        self._vocab_path = tf.saved_model.Asset(self.vocab_path)

        # Include a tokenize signature for a batch of strings.
        self.tokenize.get_concrete_function(
            inputs=tf.TensorSpec(shape=[None], dtype=tf.string),
        )

        # Include `detokenize` and `lookup` signatures for:
        #   * `Tensors` with shapes [tokens] and [batch, tokens]
        #   * `RaggedTensors` with shape [batch, tokens]
        self.detokenize.get_concrete_function(
            tf.TensorSpec(shape=[None, None], dtype=tf.int64)
        )
        self.detokenize.get_concrete_function(
            tf.RaggedTensorSpec(shape=[None, None], dtype=tf.int64)
        )

        self.lookup.get_concrete_function(
            tf.TensorSpec(shape=[None, None], dtype=tf.int64)
        )
        self.lookup.get_concrete_function(
            tf.RaggedTensorSpec(shape=[None, None], dtype=tf.int64)
        )

        # These `get_*` methods take no arguments
        self.get_vocab_size.get_concrete_function()
        self.get_vocab_path.get_concrete_function()
        self.get_reserved_tokens.get_concrete_function()

    @classmethod
    def from_pretrained(cls, model_name_or_path, *model_args, **kwargs):
        """Load tokenizer in same expected behaviour as Huggingface tokenizers."""
        tokenizer = cls(model_name_or_path, *model_args, **kwargs)
        return tokenizer

    @tf.function
    def tokenize(
        self, inputs, max_length=None, padding: Optional[PaddingStrategies] = None
    ):
        self.max_length = set_valid_max_seq_length(
            max_length, self.model_max_length, self.max_len_single_sentence
        )
        self.padding = set_valid_padding(padding)

        input_ids = self.tokenizer.tokenize(inputs)
        # Merge the `word` and `word-piece` axes.
        input_ids = input_ids.merge_dims(-2, -1)
        input_ids = self.truncate(input_ids)
        input_ids = self.post_process(input_ids)
        input_ids, mask = self.pad(input_ids)
        return {"input_ids": input_ids, "attention_mask": mask}

    def truncate(self, sequence: tf.RaggedTensor):
        """Truncation removes a sequence up to a specified lenght.

        Special tokens marking the start and end of a sequence (if specified) \
            will also be included in the sentence.
        """
        max_length = self.max_length
        return sequence[:, :max_length]

    def pad(
        self,
        sequence: tf.RaggedTensor,
    ) -> TF_TOKENIZER_INPUT:

        r"""
        Add padding of a sequence and supports different strategies for doing so.

        ::note: passing in padding will have no effect when loading a saved tokenizer.

        `strategy` supports different options:
            - `max_length`: Pad all batches to tokenizer `max_seq_length`.
            - `longest`: Pad all to same length as the longest sequence in the batch.
            - `none`: Do not pad at all.
        """
        padding = self.padding
        max_length = self.max_length
        _input_ids, _attention_mask = TF_TENSORS, TF_TENSORS  # type: ignore

        if padding == PaddingStrategies.MAX_LENGTH:
            _input_ids, _attention_mask = text.pad_model_inputs(
                input=sequence, max_seq_length=max_length
            )
        elif padding == PaddingStrategies.LONGEST:
            seq_len_per_batch: tf.Tensor = sequence.nested_row_lengths()[0]
            longest_seq_in_batch = tf.reduce_max(seq_len_per_batch)
            _input_ids, _attention_mask = text.pad_model_inputs(
                input=sequence, max_seq_length=longest_seq_in_batch
            )
        elif padding == PaddingStrategies.NONE:
            _input_ids: tf.RaggedTensor = sequence
            _attention_mask: tf.RaggedTensor = tf.ones_like(sequence)
        else:
            raise NotValidPaddingStrategy(padding=padding)

        # Since Huggingface tensors are of same type
        input_ids = tf.cast(_input_ids, dtype=tf.int64)
        attention_mask = tf.cast(_attention_mask, dtype=tf.int64)
        return input_ids, attention_mask

    @tf.function
    def post_process(self, bs: tf.RaggedTensor) -> tf.RaggedTensor:
        r"""
        Specifies how a sentence should be constructed together with a set of special tokens.

        Huggingface tokenizers specify a template for how different post_processors should behave \
            when adding special tokens to a "single" or "pair" of sentences.

        An example template parsing instruction for BERT could look like:

            >>> from tftokenizers.load import load_json
            >>> config = load_json("saved_tokenizers/bert-base-uncased/tokenizer.json")
            >>> config["post_process"]["pair"]
            \"""
            [{'SpecialToken': {'id': '[CLS]', 'type_id': 0}},
            {'Sequence': {'id': 'A', 'type_id': 0}},
            {'SpecialToken': {'id': '[SEP]', 'type_id': 0}},
            {'Sequence': {'id': 'B', 'type_id': 1}},
            {'SpecialToken': {'id': '[SEP]', 'type_id': 1}}]
            \"""

        Each type of tokenizer can have different templates between the models. \
            This function parses the template into a simple abstract syntax tree \
            and combines the original sentence with the special tokens according \
            according to the template.
        """

        num_sentences = bs.bounding_shape()[0]
        template = self.post_processor
        processing_steps = template[TemplateType.SINGLE]
        num_steps = len(processing_steps)

        def get_token_id_from_current_step_in_template(
            instruction_step: PROCESSING_STEP,
        ) -> tf.Tensor:
            token_symbol: str = list(instruction_step.values())[0]["id"]
            token_id: tf.Tensor = self.special_tokens.get(token_symbol)
            # idx = self.all_special_tokens.index(token_symbol)
            # token_id = self.all_special_ids_tf[idx]
            return token_id

        # Build AST from template instruction
        for step in range(num_steps - 1):
            left: PROCESSING_STEP = processing_steps[step]
            right: PROCESSING_STEP = processing_steps[step + 1 - num_steps]
            left_token = get_token_id_from_current_step_in_template(left)
            right_token = get_token_id_from_current_step_in_template(right)

            if left != right:
                left_sentence = bs
                right_sentence = bs

                if "SpecialToken" in left:
                    left_sentence = tf.fill([num_sentences, 1], left_token)
                elif "SpecialToken" in right:
                    right_sentence = tf.fill([num_sentences, 1], right_token)
                bs = tf.concat([left_sentence, right_sentence], axis=1)
        return bs

    def cleanup_text(self, token_txt):
        bad_tokens = [
            re.escape(tok)
            for tok in self._reserved_tokens
            if tok != SpecialTokens.UNK_TOKEN.value
        ]
        bad_token_re = "|".join(bad_tokens)

        bad_cells = tf.strings.regex_full_match(token_txt, bad_token_re)
        result = tf.ragged.boolean_mask(token_txt, ~bad_cells)

        # Join them into strings.
        result = tf.strings.reduce_join(result, separator=" ", axis=-1)
        return result

    @tf.function
    def detokenize(self, tokenized):
        words = self.tokenizer.detokenize(tokenized)
        return self.cleanup_text(words)

    @tf.function
    def lookup(self, token_ids):
        return tf.gather(self._vocab, token_ids)

    @tf.function
    def get_vocab_size(self):
        return tf.shape(self._vocab)[0]

    @tf.function
    def get_vocab_path(self):
        return self._vocab_path

    @tf.function
    def get_reserved_tokens(self):
        return tf.constant(self._reserved_tokens)

    @tf.function
    def _set_valid_max_seq_length(
        self, desired=None, model_max_length=512, num_max_tokens_in_seq=510
    ) -> int:
        """Calculate and set the maximum allowed sequence length."""
        min_length = model_max_length - num_max_tokens_in_seq
        max_length = num_max_tokens_in_seq

        if desired is None or desired > max_length or desired <= min_length:
            desired = max_length
        return desired

    @tf.function
    def _set_valid_padding(self, padding: PaddingStrategies) -> PaddingStrategies:
        padding = padding if padding is not None else PaddingStrategies.LONGEST
        return padding

    @tf.function
    def get_max_length(self):
        return self.max_length

    # New methods with names similar to
    # The Huggingface Tokenizer interface
    @tf.function
    def encode(self, strings):
        return self.tokenize(strings)

    @tf.function
    def batch_encode(self, strings):
        return self.tokenize(strings)

    @tf.function
    def batch_encode_plus(self, strings):
        return self.tokenize(strings)

    @tf.function
    def decode(self, tokenized):
        return self.detokenize(tokenized)

    @tf.function
    def batch_decode(self, tokenized):
        return self.detokenize(tokenized)


class TFTokenizerBase(tf.Module):
    # TODO: Will move to deprecate this class
    # Was based on taking in a tokenizer to start
    def __init__(
        self,
        vocab_path: str,
        tokenizer_base: text.Tokenizer,
        hf_spec: Dict[str, Any],
        config: Dict[str, Any],
        max_length: Optional[int] = None,
        padding: Optional[PaddingStrategies] = None,
    ):
        super(TFTokenizerBase, self).__init__()

        self.tokenizer = tokenizer_base
        self.config = config
        vocab = get_vocab_from_path(vocab_path)
        self.vocab_size = len(vocab)
        self.vocab = tf.Variable(vocab)

        self.model_max_length = 510  # config["model_max_length"]
        self.max_length = (
            max_length if max_length is not None else self.model_max_length
        )
        self.padding = padding if padding is not None else PaddingStrategies.LONGEST
        self.tokenizer_spec = hf_spec
        self.model = hf_spec["model"]["type"]

        self.num_special_single = num_special_tokens(TemplateType.SINGLE, hf_spec)
        self.num_special_pair = num_special_tokens(TemplateType.PAIR, hf_spec)
        special_tokens = map_special_tokens_to_ids(vocab, config)
        self.special_tokens = special_tokens
        self._reserved_tokens = list(special_tokens.keys())
        self._vocab_path = tf.saved_model.Asset(vocab_path)

        # Include a tokenize signature for a batch of strings.
        self.tokenize.get_concrete_function(
            tf.TensorSpec(shape=[None], dtype=tf.string)
        )

        # Include `detokenize` and `lookup` signatures for:
        #   * `Tensors` with shapes [tokens] and [batch, tokens]
        #   * `RaggedTensors` with shape [batch, tokens]
        self.detokenize.get_concrete_function(
            tf.TensorSpec(shape=[None, None], dtype=tf.int64)
        )
        self.detokenize.get_concrete_function(
            tf.RaggedTensorSpec(shape=[None, None], dtype=tf.int64)
        )

        self.lookup.get_concrete_function(
            tf.TensorSpec(shape=[None, None], dtype=tf.int64)
        )
        self.lookup.get_concrete_function(
            tf.RaggedTensorSpec(shape=[None, None], dtype=tf.int64)
        )

        # These `get_*` methods take no arguments
        self.get_vocab_size.get_concrete_function()
        self.get_vocab_path.get_concrete_function()
        self.get_reserved_tokens.get_concrete_function()

    @tf.function
    def tokenize(
        self, strings, max_length=None, padding: Optional[PaddingStrategies] = None
    ):
        input_ids = self.tokenizer.tokenize(strings)
        # Merge the `word` and `word-piece` axes.
        input_ids = input_ids.merge_dims(-2, -1)
        input_ids = self.truncate(input_ids)
        input_ids = self.post_process(input_ids)
        input_ids, mask = self.pad(input_ids, max_length=max_length, padding=padding)
        return {"input_ids": input_ids, "attention_mask": mask}

    def truncate(self, sequence: tf.RaggedTensor):
        """Truncation removes a sequence up to a specified lenght.

        Special tokens marking the start and end of a sequence (if specified) \
            will also be included in the sentence.
        """
        max_length = self.max_length  # TODO: update
        return sequence[:, :max_length]

    def pad(
        self,
        sequence: tf.RaggedTensor,
        max_length: int = None,
        padding: Optional[PaddingStrategies] = None,
    ) -> TF_TOKENIZER_INPUT:
        r"""
        Add padding of a sequence and supports different strategies for doing so.

        `strategy` supports different options:
            - `max_length`: Pad all batches to tokenizer `max_seq_length`.
            - `longest`: Pad all to same length as the longest sequence in the batch.
            - `none`: Do not pad at all.
        """
        _input_ids, _attention_mask = TF_TENSORS, TF_TENSORS  # type: ignore

        if padding is None:
            padding = (
                self.padding if self.padding is not None else PaddingStrategies.LONGEST
            )

        if max_length is None:
            max_length = (
                self.max_length
                if self.max_length is not None
                else self.model_max_length
            )

        if padding == PaddingStrategies.MAX_LENGTH:
            _input_ids, _attention_mask = text.pad_model_inputs(
                input=sequence, max_seq_length=max_length
            )
        elif padding == PaddingStrategies.LONGEST:
            seq_len_per_batch: tf.Tensor = sequence.nested_row_lengths()[0]
            longest_seq_in_batch = tf.reduce_max(seq_len_per_batch)
            _input_ids, _attention_mask = text.pad_model_inputs(
                input=sequence, max_seq_length=longest_seq_in_batch
            )
        elif padding == PaddingStrategies.NONE:
            _input_ids: tf.RaggedTensor = sequence
            _attention_mask: tf.RaggedTensor = tf.ones_like(sequence)
        else:
            raise NotValidPaddingStrategy(padding=padding)

        # Since Huggingface tensors are of same type
        input_ids = tf.cast(_input_ids, dtype=tf.int64)
        attention_mask = tf.cast(_attention_mask, dtype=tf.int64)
        return input_ids, attention_mask

    @tf.function
    def post_process(self, bs: tf.RaggedTensor) -> tf.RaggedTensor:
        r"""
        Specifies how a sentence should be constructed together with a set of special tokens.

        Huggingface tokenizers specify a template for how different post_processors should \
            behave when adding special tokens to a "single" or "pair" of sentences.

        An example template parsing instruction for BERT could look like:

            >>> from tftokenizers.load import load_json
            >>> config = load_json("saved_tokenizers/bert-base-uncased/tokenizer.json")
            >>> config["post_process"]["pair"]
            \"""
            [{'SpecialToken': {'id': '[CLS]', 'type_id': 0}},
            {'Sequence': {'id': 'A', 'type_id': 0}},
            {'SpecialToken': {'id': '[SEP]', 'type_id': 0}},
            {'Sequence': {'id': 'B', 'type_id': 1}},
            {'SpecialToken': {'id': '[SEP]', 'type_id': 1}}]
            \"""

        Each type of tokenizer can have different templates between the models. \
            This function parses the template into a simple abstract syntax tree \
            and combines the original sentence with the special tokens according \
            according to the template.
        """

        num_sentences = bs.bounding_shape()[0]
        template = self.tokenizer_spec["post_processor"]
        processing_steps = template[TemplateType.SINGLE]
        num_steps = len(processing_steps)

        def get_token_id_from_current_step_in_template(
            instruction_step: PROCESSING_STEP,
        ) -> tf.Tensor:
            token_symbol: str = list(instruction_step.values())[0]["id"]
            token_id: tf.Tensor = self.special_tokens.get(token_symbol)
            return token_id

        # Build AST from template instruction
        for step in range(num_steps - 1):
            left: PROCESSING_STEP = processing_steps[step]
            right: PROCESSING_STEP = processing_steps[step + 1 - num_steps]
            left_token = get_token_id_from_current_step_in_template(left)
            right_token = get_token_id_from_current_step_in_template(right)

            if left != right:
                left_sentence = bs
                right_sentence = bs

                if "SpecialToken" in left:
                    left_sentence = tf.fill([num_sentences, 1], left_token)
                elif "SpecialToken" in right:
                    right_sentence = tf.fill([num_sentences, 1], right_token)
                bs = tf.concat([left_sentence, right_sentence], axis=1)
        return bs

    def cleanup_text(self, token_txt):
        bad_tokens = [
            re.escape(tok)
            for tok in self._reserved_tokens
            if tok != SpecialTokens.UNK_TOKEN.value
        ]
        bad_token_re = "|".join(bad_tokens)

        bad_cells = tf.strings.regex_full_match(token_txt, bad_token_re)
        result = tf.ragged.boolean_mask(token_txt, ~bad_cells)

        # Join them into strings.
        result = tf.strings.reduce_join(result, separator=" ", axis=-1)
        return result

    @tf.function
    def get_max_length(self):
        return self.max_length

    @tf.function
    def detokenize(self, tokenized):
        words = self.tokenizer.detokenize(tokenized)
        return self.cleanup_text(words)

    @tf.function
    def lookup(self, token_ids):
        return tf.gather(self.vocab, token_ids)

    @tf.function
    def get_vocab_size(self):
        return tf.shape(self.vocab)[0]

    @tf.function
    def get_vocab_path(self):
        return self._vocab_path

    @tf.function
    def get_reserved_tokens(self):
        return tf.constant(self._reserved_tokens)

    # New methods with names similar to
    # The Huggingface Tokenizer interface
    @tf.function
    def encode(self, strings):
        return self.tokenize(strings)

    @tf.function
    def batch_encode(self, strings):
        return self.tokenize(strings)

    @tf.function
    def batch_encode_plus(self, strings):
        return self.tokenize(strings)

    @tf.function
    def decode(self, tokenized):
        return self.detokenize(tokenized)

    @tf.function
    def batch_decode(self, tokenized):
        return self.detokenize(tokenized)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tftokenizers']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.8.0,<2.0.0',
 'rich[jupyter]>=10.14.0,<11.0.0',
 'sentencepiece>=0.1.94,<0.2.0',
 'tensorflow-text>=2.5.0,<3.0.0',
 'tensorflow>=2.5.0,<3.0.0',
 'tf-sentencepiece>=0.1.92,<0.2.0',
 'torch>=1.8.0,<2.0.0',
 'transformers>=4.10.0,<5.0.0']

setup_kwargs = {
    'name': 'tftokenizers',
    'version': '0.1.6',
    'description': 'Use Huggingface Transformer and Tokenizers as Tensorflow Reusable SavedModels.',
    'long_description': '# TFtftransformers\n\nConverting Hugginface tokenizers to Tensorflow tokenizers. The main reason is to be able to bundle the tokenizer and model into one Reusable SavedModel, inspired by the [Tensorflow Official Guide on tokenizers](hhttps://www.tensorflow.org/text/guide/bert_preprocessing_guide)\n\n## <a href="https://badge.fury.io/py/tftokenizers"><img src="https://badge.fury.io/py/tftokenizers.svg" alt="PyPI version" height="18"></a>\n\n**Source Code**: <a href="https://github.com/Huggingface-Supporters/tftftransformers" target="_blank">https://github.com/Hugging-Face-Supporter/tftokenizers</a>\n\n---\n\nModels we know works:\n\n```python\n"bert-base-cased"\n"bert-base-uncased"\n"bert-base-multilingual-cased"\n"bert-base-multilingual-uncased"\n# Distilled\n"distilbert-base-cased"\n"distilbert-base-multilingual-cased"\n"microsoft/MiniLM-L12-H384-uncased"\n# Non-english\n"KB/bert-base-swedish-cased"\n"bert-base-chinese"\n```\n\n## Examples\n\nThis is an example of how one can use Huggingface model and tokenizers bundled together as a [Reusable SavedModel](https://www.tensorflow.org/hub/reusable_saved_models) and yields the same result as using the model and tokenizer from Huggingface 🤗\n\n```python\nimport tensorflow as tf\nfrom transformers import TFAutoModel\nfrom tftokenizers import TFModel, TFAutoTokenizer\n\n# Load base models from Huggingface\nmodel_name = "bert-base-cased"\nmodel = TFAutoModel.from_pretrained(model_name)\n\n# Load converted TF tokenizer\ntokenizer = TFAutoTokenizer.from_pretrained(model_name)\n\n# Create a TF Reusable SavedModel\ncustom_model = TFModel(model=model, tokenizer=tokenizer)\n\n# Tokenizer and model can handle `tf.Tensors` or regular strings\ntf_string = tf.constant(["Hello from Tensorflow"])\ns1 = "SponGE bob SQuarePants is an avenger"\ns2 = "Huggingface to Tensorflow tokenizers"\ns3 = "Hello, world!"\n\noutput = custom_model(tf_string)\noutput = custom_model([s1, s2, s3])\n\n# We can now pass input as tensors\noutput = custom_model(\n    inputs=tf.constant([s1, s2, s3], dtype=tf.string, name="inputs"),\n)\n\n# Save tokenizer\nsaved_name = "reusable_bert_tf"\ntf.saved_model.save(custom_model, saved_name)\n\n# Load tokenizer\nreloaded_model = tf.saved_model.load(saved_name)\noutput = reloaded_model([s1, s2, s3])\nprint(output)\n```\n\n## `Setup`\n\n```bash\ngit clone https://github.com/Hugging-Face-Supporter/tftokenizers.git\ncd tftokenizers\npoetry install\npoetry shell\n```\n\n## `Run`\n\nTo convert a Huggingface tokenizer to Tensorflow, first choose one from the models or tokenizers from the Huggingface hub to download.\n\n**NOTE**\n\n> Currently only BERT models work with the converter.\n\n### `Download`\n\nFirst download tokenizers from the hub by name. Either run the bash script do download multiple tokenizers or download a single tokenizer with the python script.\n\nThe idea is to eventually only to automatically download and convert\n\n```bash\npython tftokenizers/download.py -n bert-base-uncased\nbash scripts/download_tokenizers.sh\n```\n\n### `Convert`\n\nConvert downloaded tokenizer from Huggingface format to Tensorflow\n\n```bash\npython tftokenizers/convert.py\n```\n\n## `Before Commit`\n\n```bash\nmake build\n```\n\n## FAQ\n\n### How to know what tokenizer is used?\n**TL;DR**\n```python\nfrom transformers import AutoTokenizer\n\nname = "bert-base-cased"\ntokenizer = AutoTokenizer.from_pretrained(name)\n\n# IF the tokenizer is fast:\nprint(tokenizer.is_fast)\n# Base tokenizer model\nprint(type(tokenizer.backend_tokenizer.model))\n# Check if it is a SentencePiece tokenizer\n# Should be `vocab.txt` or `vocab.json` if not SentencePiece tokenizer\n# SencePiece if "vocab_file":\n#   "sentencepiece.bpe.model"\nprint(tokenizer.vocab_files_names)\n\n# Else\n# Find if the model is a SentencePiece model with\nprint(vars(tokenizer).get("spm_file", None))\n# print(vars(tokenizer).get("sp_model", None))\n```\n\n<details>\n<summary>:memo: Read More:</summary>\nAnd the components of the tokenizers described [here](https://huggingface.co/docs/tokenizers/python/latest/components.html) as:\n- Normalizers\n- Pre tokenizers\n- [Models](https://huggingface.co/docs/tokenizers/python/latest/components.html#models)\n- PostProcessor\n- Decoders\n\n\nWhen loading a tokenizer with Huggingface transformers, it maps the name of the model from the Huggingface Hub to the correct model and tokenizer available there, if not it will try to to find a folder on your local computer with that name.\n\nAdditionally, tokenizers from Huggingface are defined in multiple different steps using the Huggingface tokenizer library. For those interested, you can look into the different components of that library of how the composition of a tokenizer works [here](https://huggingface.co/docs/tokenizers/python/latest/). There is also a great guide documenting how composition of tokenizers are done in this [Medium article](https://towardsdatascience.com/designing-tokenizers-for-low-resource-languages-7faa4ab30ef4)\n</details>\n\n### What tokenizers are used by what models?\n<details>\n<summary>:memo: Read More:</summary>\nAs stated in the section above, you will need to look at each model to inspect the type of tokenizer it is using, but in general there are just a few "base tokenizers / models". See [Huggingface documentation](https://huggingface.co/docs/transformers/tokenizer_summary) for explanation on how these "base tokenizers" are defined\n\n[Base Tokenizer Names](https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/models/__init__.py)\n[Model Implementations](https://github.com/huggingface/tokenizers/tree/master/bindings/python/py_src/tokenizers/implementations)\n\nSentencePiece tokenizers can either be BPE (rare if the tokenizers is fast) or Unigram (all Unigram == SentencePiece)\n#### BPE = tokenizers.models.BPE\n- Implemented by\n\n    [byte-pair BPE](https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/byte_level_bpe.py), [char-level BPE](https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/char_level_bpe.py), ([SentencePiece BPE](https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/sentencepiece_bpe.py))\n\n- Used by\n\n    `GPT`, `XLNet`, `FlauBERT`, `RoBERTa`, `GPT-2`, `GPT-j`, `GPT-neo`, `BART`, `XLM-RoBERTa`\n#### Unigram = tokenizers.models.Unigram\n- Implemented by\n\n    [SentencePiece Unicode](https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/sentencepiece_unigram.py)\n\n- Used by\n\n    All `T5` models\n#### WordPiece = tokenizers.models.WordPiece\n- Implemented by\n\n    [Bert WordPiece](https://github.com/huggingface/tokenizers/blob/master/bindings/python/py_src/tokenizers/implementations/bert_wordpiece.py)\n\n- Used by\n\n    `BERT`, `mBERT`, `miniLM`, distilled versions of BERT\n\n#### SentencePiece\nSentencePiece is a method for creating sub-word tokenizations.\nIt supports BPE and Unigram.\n\nSentencePiece is a separate C++ implemented library with python and Tensorflow bindings.\nThe vocabulary is bundled into:\n\n**For fast models**:\n\n"vocab_file_names":\n\n    `sentencepiece.bpe.model` for "BPE" and\n    `spiece.model` for Unigram\n\n**For slow models**:\n\n"vocab_file_names":\n\n    \'source_spm\': \'source.spm\',\n    \'target_spm\': \'target.spm\',\n    \'vocab\': \'vocab.json\'\n\n"spm_files":\n\n    will be a single file or a list of files\n    ...\n\n- Used by:\n\n    **Fast**: `T5` models\n    **Slow**: `facebook/m2m100_418M`, `facebook/wmt19-en-de`\n</details>\n\n### How to implement the tokenizers from Huggingface to Tensorflow?\nYou will need to download the Huggingface tokenizer of your choice, determine the type of the tokenizer (`is_fast`, tokenizer type and `vocab_file_names`). Then map the tokenizer used to the Tensorflow supported equivalent:\n\nhttps://github.com/tensorflow/text/issues/422\n\n**BPE** and **Unigram**:\n- All BPE implementations for Tensorflow is backed by SentencePiece\n- [SentencePiece in TensorFlow](https://www.tensorflow.org/text/api_docs/python/text/SentencepieceTokenizer)\n- [Official Answer 1](https://github.com/tensorflow/text/issues/415)\n- [Official Answer 2](https://github.com/tensorflow/text/issues/763)\n- [How to load a SentencePiece model](https://github.com/tensorflow/text/issues/215)\n- [Input will need to be Tensors](https://github.com/tensorflow/text/issues/512)\n- [How to load model from vocab](https://github.com/tensorflow/text/issues/452)\n\n**WordPiece**:\n- [BertTokenizer](https://www.tensorflow.org/text/api_docs/python/text/BertTokenizer) or\n- [WordPiece](https://www.tensorflow.org/text/api_docs/python/text/FastWordpieceTokenizer) or\n- [FastWordPiece](https://www.tensorflow.org/text/api_docs/python/text/FastWordpieceTokenizer)\n\n\nhttps://github.com/tensorflow/text/issues/116\nhttps://github.com/tensorflow/text/issues/414\n\n### What other ways are there to convert a tokenizer?\n<details>\n<summary>:memo: Read More:</summary>\nWith `tfokenizers` there are three ways to use the package:\n\n```python\nimport tensorflow as tf\nimport tensorflow_text as text\nfrom transformers import AutoTokenizer, TFAutoModel\nfrom transformers.utils.logging import set_verbosity_error\n\nfrom tftokenizers.file import (\n    get_filename_from_path,\n    get_vocab_from_path,\n    load_json\n)\nfrom tftokenizers.model import TFModel\nfrom tftokenizers.tokenizer import TFAutoTokenizer, TFTokenizerBase\n\nset_verbosity_error()\ntf.get_logger().setLevel("ERROR")\n\npretrained_model_name = "bert-base-cased"\n\n\n# a) by model_name\ntf_tokenizer = TFAutoTokenizer.from_pretrained(pretrained_model_name)\n\n# b) bundled with the model, similar to TFHub\nmodel = TFAutoModel.from_pretrained(pretrained_model_name)\ncustom_model = TFModel(model=model, tokenizer=tf_tokenizer)\n\n# c) from source, using the saved files of a transformers tokenizer\n# Make sure you run download.py or the download script first\nPATH = "saved_tokenizers/bert-base-uncased"\nvocab = get_vocab_from_path(PATH)\nvocab_path = get_filename_from_path(PATH, "vocab")\n\nconfig = load_json(f"{PATH}/tokenizer_config.json")\ntokenizer_spec = load_json(f"{PATH}/tokenizer.json")\nspecial_tokens_map = load_json(f"{PATH}/special_tokens_map.json")\n\ntokenizer_base_params = dict(lower_case=True, token_out_type=tf.int64)\ntokenizer_base = text.BertTokenizer(vocab_path, **tokenizer_base_params)\ncustom_tokenizer = TFTokenizerBase(\n    vocab_path=vocab_path,\n    tokenizer_base=tokenizer_base,\n    hf_spec=tokenizer_spec,\n    config=config,\n)\n```\n</details>\n\n\n### How to save Huggingface Tokenizer files locally?\n<details>\n<summary>:memo: Read More:</summary>\n\nTo download the files used by Huggingface tokenizers, you can either download one by name\n```\npython tftokenizers/download.py -n KB/bert-base-swedish-cased\n```\nor download multiple\n```\nbash scrips/download_tokenizers.sh\n```\n</details>\n\n## WIP\n\n- [x] Convert a BERT tokenizer from Huggingface to Tensorflow\n- [x] Make a TF Reusabel SavedModel with Tokenizer and Model in the same class. Emulate how the TF Hub example for BERT works.\n- [x] Find methods for identifying the base tokenizer model and map those settings and special tokens to new tokenizers\n- [x] Extend the tokenizers to more tokenizer types and identify them from a huggingface model name\n- [x] Document how others can use the library and document the different stages in the process\n- [x] Improve the conversion pipeline (s.a. Download and export files if not passed in or available locally)\n- [ ] `model_max_length` should be regulated. However, some newer models have the max_lenght for tokenizers at 1000_000_000\n- [ ] Support more tokenizers, starting with SentencePiece\n- [ ] Identify tokenizer conversion limitations\n- [ ] Support encoding of two sentences at a time [Ref](https://www.tensorflow.org/text/guide/bert_preprocessing_guide)\n- [ ] Allow the tokenizers to be used for Masking (MLM) [Ref](https://www.tensorflow.org/text/guide/bert_preprocessing_guide)\n',
    'author': 'MarkusSagen',
    'author_email': 'markus.john.sagen@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Hugging-Face-Supporter/tftokenizers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
